from typing import Union


class PyGraphVisual:
    """
    Facilitate data preparation for graph visualization using the Cytoscape.js library
    """


    def __init__(self, db=None):
        self.db = db                    # Object of "NeoAccess" class

        self.structure = []             # The data that defines a graph to visualize.
                                        #   A list of dicts defining nodes, and possibly edges as well.
                                        #   NODES must contain the keys 'id' and 'labels'
                                        #   EDGES must contain the keys 'name', 'source', and 'target'
                                        #       (and presumably 'id' is required as well?)
                                        # EXAMPLE (2 nodes and an edge):
                                        #   [{'id': 1, 'labels': ['PERSON'], 'name': 'Julian'},
                                        #    {'id': 2, 'labels': ['CAR'], 'color': 'white'},
                                        #    {'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}]

        self.color_mapping = {}         # Mapping a node label to its interior color (the edge color is an automatic variation)
                                        # EXAMPLE:  {'PERSON': 'cyan', 'CAR': 'orange'}

        self.caption_mapping = {}       # Mapping a node label to its caption (field name to use on the graph)
                                        # EXAMPLE:  {'PERSON': 'name', 'CAR': 'color'}

        self._next_available_edge_id = 1 # An auto-increment value

        self._all_node_ids = []         # List of all the node id's added to the graph so far;
                                        #   used for optional prevention of duplicates


    def serialize(self) -> dict:
        """
        Extract and return a dict of relevent data from this object

        :return:
        """
        return {"structure": self.structure,
                "color_mapping": self.color_mapping,
                "caption_mapping": self.caption_mapping}



    def __str__(self) -> str:
        """
        Return an overview description of this object

        :return:    A string with a description of this object
        """
        s = f"Object describing a graph containing a total of {len(self.structure)} nodes and edges:\n"
        s += f"    Graph structure: {self.structure}\n"
        s += f"    Graph color mapping: {self.color_mapping}\n"
        s += f"    Graph caption mapping: {self.caption_mapping}"

        return s



    def get_graph_data(self) -> dict:
        """
        Return a data dictionary with all the relevant visualization info
        (typically, to pass to the front-end, for eventual use by the Cytoscape.js library)

        :return:    A dict with 3 keys - "structure", "color_mapping" and "caption_mapping"
        """
        return {"structure": self.structure,
                "color_mapping": self.color_mapping,
                "caption_mapping": self.caption_mapping}



    def add_node(self, node_id :Union[int,str], labels="", data=None) -> None:
        """
        Prepare and store the data for 1 node, in a format expected by the visualization front-end

        EXAMPLE:    {'id': 1, 'name': 'Julian', 'labels': ['PERSON']}

        :param node_id:         Either an integer or a string to uniquely identify this node in the graph;
                                    it will be used to specify the start/end nodes of edges.
                                    Typically, use either the internal database ID, or some URI
                                    Records with a duplicate node_id are silently disregarded
        :param labels:          A string, or list of strings, with the node's label(s) used in the graph database;
                                    if not used, pass an empty string
        :param data:            A dict with all other node data not already specified in any of the other arguments
        :return:                None
        """
        assert node_id != "", \
            "add_node(): cannot use an empty string for the argument `node_id`"


        if node_id in self._all_node_ids:
            return      # Silently disregard duplicates

        if data is None:
            data = {}

        d = data.copy()     # Make an independent clone of the data dict


        d["id"] = node_id

        if type(labels) == str:
            labels = [labels]

        d["labels"] = labels

        self.structure.append(d)
        self._all_node_ids.append(node_id)



    def add_edge(self, from_node :Union[str, int], to_node :Union[str, int],
                 name :str, edge_id=None, data=None) -> None:
        """
        Prepare and store the data for 1 edge, in a format expected by the visualization front-end.

        EXAMPLE:   {'name': 'OWNS', 'source': 1, 'target': 2, 'id': 'edge-1'}

        Note that 'id', in this context, is whatever we pass to the visualization module for the nodes;
        not necessarily related to the internal database ID's

        :param from_node:   Integer or a string uniquely identify the "id" of the node where the edge originates
        :param to_node:     Integer or a string uniquely identify the "id" of the node where the edge ends
        :param name:        Name of the relationship
        :param edge_id:     (OPTIONAL)  If not provided, strings such as "edge-123" are used,
                                        with auto-generated consecutive integers
                                TODO: unclear if edge id's are really needed
        :param data:        A dict with all other node data not already specified in any of the other arguments

        :return:            None
        """
        from_id = from_node
        to_id = to_node

        d = {"name": name, "source": from_id, "target": to_id}

        if edge_id is not None:
            d["id"] = edge_id
        else:
            d["id"] = f"edge-{self._next_available_edge_id}"
            self._next_available_edge_id += 1        # Maintain an auto-increment value

        if data:
            d.update(data)

        self.structure.append(d)



    def assign_caption(self, label :str, caption="name") -> None:
        """
        Assign and store a mapping from label name to caption (name of field to use on the display)

        EXAMPLES:   assign_caption(label='PERSON', caption='name')
                    assign_caption(label='CAR', caption='color')

        :param label:   The name of a node label
        :param caption: The name of the node field to use on the display (by default, "name")
        :return:        None
        """
        assert type(label) == str, \
            "assign_caption(): the argument `label` must be a string"

        self.caption_mapping[label] = caption



    def assign_color_mapping(self, label :str, color :str) -> None:
        """
        Assign and store a mapping from label name to color to use for nodes having that label

        :param label:   The name of a node label
        :param color:   The name (hex string or standard CSS name) of the color to use on the display
        :return:        None
        """
        extra_colors =  {       # Convenient extra colors, not available thru standard CSS names
            "graph_green": '#8DCC92',
            "graph_darkgreen": '#56947E',
            "graph_teal": '#569481',
            "graph_orange": '#F79768',
            "graph_blue": '#4C8EDB',
            "graph_red": '#F16668',
            "graph_darkbrown": '#604A0D',
            "graph_lightbrown": '#D9C8AD',
            "graph_orchid": '#C990C1',
            "graph_gold": '#FFC455'     # Has a bit more orange than 'gold'
        }

        if color in extra_colors:
            color = extra_colors[color]

        self.color_mapping[label] = color




    ##########   The method below require a database connection   ##########


    def prepare_graph(self, result_dataset :[dict], add_edges=True) -> [int]:
        """
        Given a list of dictionary data about graph-database nodes - for example,
        as returned by NeoAccess.get_nodes() - construct and save visualization data for them.

        Each dictionary entry is expected to have a key named "internal_id";
        if not present, it will be silently ignored.
        Though not required, a key named "neo4j_labels" is typically present as well.

        :param result_dataset:  A list of dictionary data about graph-database nodes
        :param add_edges:       If True, all existing edges among the displayed nodes
                                    will also be part of the visualization
        :return:                A list of integers with the internal databased IDs
                                    of all the nodes added to the graph structure
        """
        node_list = []      # Running list of internal databased IDs, for nodes included in visualization
        for node in result_dataset:
            internal_id = node.get("internal_id")
            if not internal_id:
                continue    # Silently ignore any entry lacking this data

            node_list.append(internal_id)

            if "neo4j_labels" in node:
                labels = node["neo4j_labels"]
                del node["neo4j_labels"]
            else:
                labels = ""

            self.add_node(node_id=internal_id, labels=labels,
                          data=node)


        if add_edges:
            # Search the database for any edges among any of the nodes selected for the visualization
            q = '''
                MATCH (n1)-[r]->(n2) 
                WHERE ID(n1) IN $node_list AND ID(n2) IN $node_list 
                RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, type(r) AS name
                '''

            result = self.db.query(q, {"node_list": node_list})
            for edge in result:
                #print(edge)
                self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"], name=edge["name"])

        return node_list



    def link_node_groups(self, group1 :[int], group2 :[int]) -> None:
        """
        Search the database for any edges from any of the nodes in the 1st group, to any node in the 2nd group.
        Any located edge will be added to the visualization data stored in this object

        :param group1:  List of integers with the internal databased IDs of the 1st group of nodes
        :param group2:  List of integers with the internal databased IDs of the 2nd group of nodes
        :return:        None
        """
        q = '''
            MATCH (n1)-[r]->(n2) 
            WHERE ID(n1) IN $group1 AND ID(n2) IN $group2 
            RETURN DISTINCT id(n1) AS from_node, id(n2) AS to_node, type(r) AS name
            '''

        result = self.db.query(q, {"group1": group1, "group2": group2})
        for edge in result:
            #print(edge)
            self.add_edge(from_node=edge["from_node"], to_node=edge["to_node"], name=edge["name"])



import numpy as np
import pandas as pd
from typing import Union
from life123.collections import CollectionTabular



class Diagnostics:
    """
    For the management of reaction diagnostic data
    """

    def __init__(self, chem_data):

        assert chem_data is not None, \
            "Diagnostics class cannot be instantiated with a missing value for argument `chem_data`"

        self.chem_data = chem_data

        # TODO: maybe drop the "diagnostic_" from the names, or rename it to "historic_"
        self.diagnostic_conc_data = CollectionTabular(parameter_name="TIME")
                                        # An expanded version of the normal System History.
                                        #   Columns of the dataframes:
                                        #       'TIME' 	'A' 'B' ...  'caption'
                                        #
                                        #   Note: if an interval run is aborted, NO entry is created here
                                        #         (this approach DIFFERS from that of other diagnostic data)


        self.diagnostic_rxn_data = {}   # "Diagnostic reaction data", PER REACTION: a dict with as many entries as reactions.
                                        #   The keys are the reaction indices; the values are objects of type "MovieTabular",
                                        #   which contain Pandas dataframes with the following columns
                                        #   (referring to one specific reaction):
                                        #           'START TIME' , 'time_step' , 'aborted', 'Delta A' , 'Delta B' , ... , 'rate', 'caption'
                                        #
                                        #   Note:   - entries are always added, even if a time-step run is aborted
                                        #           - the various 'Delta concentrations' are for the chemicals involved in the reaction
                                        #             over the time interval that *STARTS* at the value in the "START TIME" column


        self.diagnostic_decisions_data = CollectionTabular(parameter_name="START_TIME")
                                        #   Columns of the dataframes:
                                        #       'START_TIME' 	'Delta A' 'Delta B' ...
                                        #               [plus, if applicable, other fields such as
                                        #               'action', 'norm_A', 'norm_B', 'step_factors']
                                        #
                                        #   Note: entries are always added, even if an interval run is aborted




    #####  1. diagnostic_rxn_data  #####

    def save_rxn_data(self, rxn_index :int, system_time, time_step,
                      increment_vector_single_rxn: Union[np.array, None],
                      aborted=False,
                      rate=None,
                      caption="") -> None:
        """
        Save up diagnostic data for 1 reaction, for a simulation step
        (by convention, regardless of whether the step is completed or aborted)

        :param rxn_index:                   The integer index (0-based) to identify the reaction of interest
        :param system_time:                 The START time of the reaction step
        :param time_step:                   The duration of the current simulation step
        :param increment_vector_single_rxn: A Numpy array of size equal to the total number of chemical species,
                                                containing the "delta concentrations" caused by this reaction
                                                for ALL the chemicals (whether involved in the reaction or not);
                                                it may be None if the current reaction step was aborted
                                                Note: diagnostic data is saved only for the chemicals involved in this reaction
        :param aborted:                     True is the current reaction step was aborted (i.e. will get repeated);
                                                False (default) for normal situations
        :param rate:                        [OPTIONAL] The value of the reaction rate (aka reaction velocity) at this step
        :param caption:                     [OPTIONAL] string to describe the snapshot
        :return:                            None
        """
        # Validate the reaction index
        self.chem_data.assert_valid_rxn_index(rxn_index)

        # Sorted list of the indexes of all the chemicals participating in this reaction
        indexes = self.chem_data.get_chemicals_indexes_in_reaction(rxn_index)

        # Validate increment_vector_single_rxn
        if increment_vector_single_rxn is not None:     # If the values are available (not the case in aborts)
            assert len(increment_vector_single_rxn) == self.chem_data.number_of_chemicals(), \
                f"save_diagnostic_rxn_data(): the length of the Numpy array increment_vector_single_rxn " \
                f"({len(increment_vector_single_rxn)}) doesn't match the overall number of chemical ({self.chem_data.number_of_chemicals()})"


        # Initialize a "MovieTabular" object for this reaction, if needed
        if rxn_index not in self.diagnostic_rxn_data:
            self.diagnostic_rxn_data[rxn_index] = CollectionTabular(parameter_name="START_TIME")


        data_snapshot = {"time_step": time_step, "aborted": aborted}      # Dict being prepared to add a new row to a Pandas dataframe

        if increment_vector_single_rxn is None:     # If the values aren't available (as is the case in aborts)
            for index in indexes:       # For all the chemicals participating in this reaction
                data_snapshot["Delta " + self.chem_data.get_label(index)] = np.nan
        else:
            for index in indexes:       # For all the chemicals participating in this reaction
                data_snapshot["Delta " + self.chem_data.get_label(index)] = increment_vector_single_rxn[index]

        if rate is not None:
            data_snapshot["rate"] = rate

        self.diagnostic_rxn_data[rxn_index].store(par=system_time,
                                                  data_snapshot=data_snapshot, caption=caption)



    def save_diagnostic_aborted_rxns(self, system_time, time_step,
                                    caption="") -> None:
        """
        Save up diagnostic data, for ALL reactions, in case of aborted simulation step

        :param system_time: The START time of the reaction step
        :param time_step:   The duration of the current simulation step
        :param caption:     [OPTIONAL] string to describe the snapshot
        :return:            None
        """
        for rxn_index in range(self.chem_data.number_of_reactions()):
            self.save_rxn_data(system_time=system_time, time_step=time_step,
                               increment_vector_single_rxn=None,
                               aborted=True,
                               rxn_index=rxn_index, caption=caption)



    def annotate_abort_rxn_data(self, msg: str) -> None:
        """
        Store the given value in the caption field of the last record
        for EACH of the reaction-specific dataframes.
        Also, set the "aborted" fields to "True"

        :param msg: Value to set the comment field to
        :return:    None
        """
        for rxn_index in range(self.chem_data.number_of_reactions()):
            self.diagnostic_rxn_data[rxn_index].set_caption_last_snapshot(msg)
            self.diagnostic_rxn_data[rxn_index].set_field_last_snapshot(field_name="aborted", field_value="True")
            # TODO: investigate why using "True" rather then True



    def get_system_history_with_rxn_rates(self, rxn_index :int) -> pd.DataFrame:
        """
        Return a Pandas dataframe containing the saved System History,
        plus an extra column named "rate", for the reaction rates of the specified reaction.
        Notice: the last row of the System History does NOT get included,
                because no rate information is known about the next simulation step not taken

        :param rxn_index:   The integer index (0-based) to identify the reaction of interest
        :return:            A Pandas data frame with the following columns:
                                "TIME"
                                A series of columns for each of the registered chemicals
                                "caption"
                                "rate"
        """
        # TODO: probably ditch, because no longer needed.  (Also, the dataframe merge might not always work...)
        # TODO: the times had better match up!  Currently not validated

        rates = self.get_rxn_rates(rxn_index=rxn_index)     # A Pandas dataframe with 2 columns: "START_TIME" and "rate"
        system_history = self.get_diagnostic_conc_data()    # Note that is a copy of the dataframe; so, no harm in changing it, below
        # Dropping the last row, because no rate information is known about the next simulation step not taken!
        system_history = system_history[:-1]

        # They'd better match up!
        assert len(system_history) == len(rates), \
            f"get_system_history_with_rxn_rates() : unable to reconcile " \
            f"the system history data with the reaction data for the requested reaction (index {rxn_index}). " \
            f"The diagnostic concentration data (excluding the last entry, which can't have a rate value) contains {len(system_history)} records, " \
            f"while the diagnostic rates data contains {len(rates)} values"

        system_history["rate"] =  rates["rate"]

        return system_history



    def get_rxn_rates(self, rxn_index :int) -> Union[pd.DataFrame, None]:
        """
        Return a Pandas dataframe with 2 columns: "START_TIME" and "rate",
        with the rates of the specified reaction at those times.

        Note that the dataframe will have 1 LESS ROW than a corresponding dataframe of concentration history,
        because the rate at the final time won't be known (since that time would be the start of a simulation step
        that has not been done)

        Also, aborted time steps are eliminated (only the rate of the final "re-do" is included)

        :param rxn_index:   The integer index (0-based) to identify the reaction of interest
        :return:            A Pandas dataframe with times and their corresponding reaction rates
        """
        rxn_df = self.get_rxn_data(rxn_index=rxn_index)
        if rxn_df is None:
            return None

        # Ditch any row where the "aborted" field has a True value; only take 2 columns of the resulting dataframe
        df_filtered = rxn_df[rxn_df['aborted'] == False][["START_TIME", "rate"]]

        return df_filtered.reset_index(drop=True)   # Re-number rows (drop=True avoids adding the old index as a new column)



    def get_rxn_data(self, rxn_index :int, head=None, tail=None,
                     t=None, print_reaction=True) -> Union[pd.DataFrame, None]:
        """
        Return a Pandas dataframe with the diagnostic run data of the requested SINGLE reaction,
        from the time that the diagnostics were enabled by instantiating this class.

        In particular, the dataframe contains the reaction rate at the start time,
        and the "Delta" values for each of the chemicals
        involved in the reaction - i.e. the change in their concentrations
        over the time interval that *STARTS* at the value in the "START_TIME" column.
        (So, there'll be no row with the final current System Time)

        Note: entries are always added, even if an interval run is aborted and automatically re-done;
              therefore, some entries will be duplicates in the "START_TIME" column.
              Those entries with have a value of True in the "aborted" column.

        Optionally, print out a brief description of the reaction.

        Optionally, limit the dataframe to a specified numbers of rows at the end,
        or just return one entry corresponding to a specific time
        (the row with the CLOSEST time to the requested one, which will appear in an extra column
        called "search_value")

        :param rxn_index:       The integer index (0-based) to identify the reaction of interest

        :param head:            (OPTIONAL) Number of records to return,
                                    from the start of the diagnostic dataframe.
        :param tail:            (OPTIONAL) Number of records to return,
                                    from the end of the diagnostic dataframe.
                                    If either the "head" arguments is passed, this argument will get ignored

        :param t:               (OPTIONAL) Individual time to pluck out from the dataframe;
                                    the row with closest time will be returned.
                                    If this parameter is specified, an extra column - called "search_value" -
                                    is inserted at the beginning of the dataframe
                                    If either the "head" or the "tail" arguments are passed, this argument will get ignored

        :param print_reaction:  (OPTIONAL) If True (default), concisely print out the requested reaction

        :return:                If present, return a Pandas data frame with (all or some of)
                                    the diagnostic data of the specified reaction.
                                    Columns of the dataframes:
                                    'START_TIME','time_step','aborted','Delta A','Delta B'...,'rate','caption'
                                If not present, return None
        """
        # Validate the reaction index
        self.chem_data.assert_valid_rxn_index(rxn_index)

        if print_reaction:
            print("Reaction: ", self.chem_data.single_reaction_describe(rxn_index=rxn_index, concise=True))

        movie_obj = self.diagnostic_rxn_data.get(rxn_index)    # Object of type "MovieTabular"

        if movie_obj is None:
            print(f"get_diagnostic_rxn_data(): no diagnostics data exists for reaction index {rxn_index} ; "
                  f"did you run the reaction simulation prior to calling this function?" )
            return None


        return movie_obj.get_dataframe(head=head, tail=tail, search_col="START_TIME", search_val=t, return_copy=True)




    #####  2. diagnostic_conc_data  #####

    def save_diagnostic_conc_data(self, system_data, system_time, caption="") -> None:
        """
        Save the diagnostic concentration data during the run, indexed by the current System Time.
        Note: if an interval run is aborted, by convention NO entry is created here

        :return: None
        """
        self.diagnostic_conc_data.store(par=system_time,
                                        data_snapshot=system_data, caption=caption)



    def get_diagnostic_conc_data(self) -> pd.DataFrame:
        """
        Return the diagnostic concentration data saved during the run.
        This will be a complete set of simulation steps,
        even if the user elected to only save part of the history during the run.

        Note: if the run of an interval step is aborted,
              by convention NO entry is created here

        :return: A Pandas dataframe, with the columns:
                    'TIME' 	'A' 'B' ... 'caption'
                    where 'A', 'B', ... are the labels of all the chemicals
        """
        return self.diagnostic_conc_data.get_dataframe(return_copy=True)




    #####  3. diagnostic_decisions_data  #####

    def save_diagnostic_decisions_data(self, system_time, data :dict,
                                       delta_conc_arr :Union[np.ndarray, None], caption="") -> None:
        """
        Used to save the diagnostic concentration data during the run, indexed by the given System Time.
        Note: if an interval run is aborted, by convention an entry is STILL created here

        :param system_time:
        :param data:
        :param delta_conc_arr:  A Numpy array of "delta concentrations".  EXAMPLE: array[1.23, 52.2]
        :param caption:         [OPTIONAL] String with a caption for this record
        :return:                None
        """
        delta_conc_dict = {}
        if delta_conc_arr is not None:
            delta_conc_dict = self._delta_conc_dict(delta_conc_arr)
            # EXAMPLE:  {"Delta A": 1.23, "Delta X": 52.2}


        delta_conc_dict.update(data)        # Merge the data dict into the delta_conc_dict

        self.diagnostic_decisions_data.store(par=system_time,
                                             data_snapshot=delta_conc_dict, caption=caption)



    def get_decisions_data(self) -> pd.DataFrame:
        """
        Determine and return the diagnostic data about concentration changes at every step - by convention,
        EVEN aborted ones
    
        :return:    A Pandas dataframe with a "TIME" column, and columns for all the "Delta concentration" values
        """
        return self.diagnostic_decisions_data.get_dataframe(return_copy=True)




    #############  EXPLAIN THINGS  #############


    def explain_reactions(self) -> bool:
        """
        Provide a detailed explanation of all the steps of the reactions,
        from the saved diagnostic data

        WARNING: Currently designed only for exactly 2 reactions!  TODO: generalize to any number of reactions

        TODO: test and validate usefulness, now that substeps got eliminated
        TODO: allow arguments to specify the min and max reaction times during which to display the explanatory data

        :return:    True if the diagnostic data is consistent for all the steps of all the reactions,
                    or False otherwise
        """
        number_reactions = self.chem_data.number_of_reactions()

        assert number_reactions == 2, \
            "explain_reactions() currently ONLY works when exactly 2 reactions are present. " \
            "Future versions will lift this restriction"

        row_baseline = 0
        row_list = [0, 0]       # TODO: generalize
        active_list = [0, 1]    # ALL the reactions.  TODO: generalize

        self._explain_reactions_helper(active_list=active_list,
                                       row_baseline=row_baseline, row_list=row_list)


        while row_baseline < len(self.get_diagnostic_conc_data()) - 2:
            row_baseline += 1

            print("Advance a single-step across all tables")

            for i in active_list:
                row_list[i] += 1

            status = self._explain_reactions_helper(active_list=active_list, row_baseline=row_baseline, row_list=row_list) # TODO: generalize


            if not status:
                return False    # Error termination

        # END while

        return True             # Successful termination



    def _explain_reactions_helper(self, active_list, row_baseline, row_list) -> bool:
        """
        Helper function for explain_reactions()

        :param active_list:
        :param row_baseline:
        :param row_list:
        :return:            True is the diagnostic data is consistent for this step, or False otherwise
        """
        print("-----------")
        print("ROW of baseline data: ", row_baseline)

        current_time = self.get_diagnostic_conc_data().loc[row_baseline]['TIME']
        print(f"TIME = {current_time:.5g}")

        print("row_list: ", row_list)
        print("active_list: ", active_list)

        chemical_list = self.chem_data.get_all_labels()
        chemical_delta_list = self._delta_names()

        conc_arr_before = self.get_diagnostic_conc_data().loc[row_baseline][chemical_list].to_numpy(dtype='float16')
        print("baseline concentrations: ", conc_arr_before)

        delta_cumulative = np.zeros(self.chem_data.number_of_chemicals(),
                                    dtype=float)  # One element per chemical species

        # For each reaction
        for rxn_index in range(self.chem_data.number_of_reactions()):
            if (rxn_index in active_list):  # If reaction is tagged as "fast"
                row = row_list[rxn_index]   # Row in the data frame for the diagnostic data on this reaction
                delta_rxn = self.get_rxn_data(rxn_index=rxn_index).loc[row][chemical_delta_list].to_numpy(dtype='float16')
                print(f"From fast rxn {rxn_index}: delta_rxn = {delta_rxn}")
            else:                           # If reaction is tagged as "slow"
                row = row_list[rxn_index]   # Row in the data frame for the diagnostic data on this reaction
                delta_rxn = self.get_rxn_data(rxn_index=rxn_index).loc[row][chemical_delta_list].to_numpy(dtype='float16')
                #delta_rxn = np.zeros(self.reaction_data.number_of_chemicals(),
                #                     dtype=float)   # This is the way it was in release beta 19
                print(f"From slow rxn {rxn_index}: delta_rxn = {delta_rxn}")


            delta_cumulative += delta_rxn
        # END for

        print("delta_cumulative: ", delta_cumulative)

        conc_after = conc_arr_before + delta_cumulative
        print("updated concentrations: ", conc_after)

        next_system_state = self.get_diagnostic_conc_data().loc[row_baseline + 1][chemical_list].to_numpy(dtype='float16')
        print(f"concentrations from the next row ({row_baseline + 1}) of the system state: ", next_system_state)

        status = np.allclose(conc_after.astype(float), next_system_state.astype(float))
        if status:
            print("Match OK")
        else:
            print("****************************************   MISMATCH!!!  ****************************************")

        print("-----------")

        return status



    def explain_time_advance(self, return_times=False, silent=False, sys_history=None) -> Union[None, tuple]:
        """
        Use the saved-up diagnostic data, to print out details of the timescales of the reaction run

        If diagnostics weren't enabled ahead of calling this function, an Exception is raised

        EXAMPLE of output:
            From time 0 to 0.0304, in 17 FULL steps of 0.0008
            (for a grand total of 38 FULL steps)

        :param return_times:    If True, all the critical times (times where the interval steps change)
                                    are saved and returned as a list
        :param silent:          If True, nothing gets printed out
        :param sys_history:     [OPTIONAL] The system history.
                                    If passed, use it in lieu of the diagnostic data;
                                    a consideration is the fact that the user might only have asked
                                    for PART of the history to be saved
        :return:                Depending on the argument "return_times", it returns either None, or a pair with 2 lists:
                                        1 - list of time values
                                        2 - list of step sizes  (will have one less element than the first list)
        """
        if sys_history:
            df = sys_history
            assert not df.empty , \
                "Diagnostics.explain_time_advance(): no history data found.  " \
                "Did you run the reaction simulation prior to calling this function?"
        else:
            df = self.get_diagnostic_conc_data()
            assert not df.empty , \
                "Diagnostics.explain_time_advance(): no diagnostic data found.  " \
                "Did you run the reaction simulation prior to calling this function?"

        n_entries = len(df)
        if sys_history:
            t = list(df["SYSTEM TIME"])     # List of times (simulation step points)
        else:
            t = list(df["TIME"])            # List of times (simulation step points)

        if n_entries == 1:
            print(f"Diagnostics.explain_time_advance(): no time advance found. "
                  f"Diagnostics only contain an initial System Time of {t[0]:.3g} ;  "
                  f"did you run the reaction simulation prior to calling this function?")
            return


        # If we get thus far, we have at least 2 entries in the list "t" of times
        grad = np.diff(t)
        grad_shifted = np.insert(grad, 0, 0)  # Insert a zero at the top (shifting down the array)
        #print(grad)
        #print(grad_shifted)
        start_interval = t[0]

        critical_times = [start_interval]  # List of times to report on (start time, plus times where the steps change)
        step_sizes = []

        total_number_full_steps = 0
        for i in range(1, n_entries-1):
            #print(i)
            if not np.allclose(grad[i], grad_shifted[i]):
                #print (f"   Detection at element {i} : time={t[i]}")
                #print (f"   From time {start_interval} to {t[i]}, in steps of {grad_shifted[i]}")
                n_full_steps_taken = self._explain_time_advance_helper(t_start=start_interval, t_end=t[i],
                                                                       delta_baseline=grad_shifted[i], silent=silent)

                start_interval = t[i]
                if n_full_steps_taken > 0:
                    total_number_full_steps += n_full_steps_taken
                    critical_times.append(t[i])
                    step_sizes.append(grad_shifted[i])


        # Final wrap-up of the interval's endpoint
        n_full_steps_taken = self._explain_time_advance_helper(t_start=start_interval, t_end=t[-1],
                                                               delta_baseline=grad_shifted[-1], silent=silent)

        if n_full_steps_taken > 0:
            total_number_full_steps += n_full_steps_taken
            critical_times.append(t[-1])
            step_sizes.append(grad_shifted[-1])

        if not silent:
            print(f"({total_number_full_steps} steps total)")

        if return_times:
            assert len(critical_times) == len(step_sizes) + 1, \
                "explain_time_advance(): validation error in the values to return"
            return (critical_times, step_sizes)



    def _explain_time_advance_helper(self, t_start, t_end, delta_baseline, silent: bool) -> Union[int, float]:
        """
        Using the provided data, about a group of same-size steps, create and print a description of it for the user

        :param t_start:
        :param t_end:
        :param delta_baseline:
        :param silent:          If True, nothing gets printed; otherwise, a line is printed out
        :return:                The corresponding number of FULL steps taken
        """
        if np.allclose(t_start, t_end):
            #print(f"   [Ignoring interval starting and ending at same time {t_start:.3g}]")
            return 0

        n_steps = round((t_end - t_start) / delta_baseline)
        step_s = "step" if n_steps == 1 else "steps"  # singular vs. plural
        if not silent:
            print(f"From time {t_start:.4g} to {t_end:.4g}, in {n_steps} {step_s} of {delta_baseline:.3g}")
        return n_steps



    def stoichiometry_checker(self, rxn_index :int, conc_arr_before: np.array, conc_arr_after: np.array) -> bool:
        """
        For the indicated reaction, investigate the change in the concentration of the involved chemicals,
        to ascertain whether the change is consistent with the reaction's stoichiometry.
        See https://life123.science/reactions

        IMPORTANT: this function is currently meant for simulations involving only 1 reaction (TODO: generalize)

        NOTE: the concentration changes in chemicals not involved in the specified reaction are ignored

        :param rxn_index:       The integer index (0-based) to identify the reaction of interest
        :param conc_arr_before: Numpy array with the concentrations of ALL the chemicals (whether involved
                                    in the reaction or not), in their index order, BEFORE the reaction step
        :param conc_arr_after:  Same as above, but after the reaction
                                TODO: maybe also accept a Panda's dataframe row

        :return:                True if the change in reactant/product concentrations is consistent with the
                                    reaction's stoichiometry, or False otherwise
        """
        assert self.chem_data.number_of_reactions() == 1, \
            f"stoichiometry_checker() currently only works for 1-reaction simulations, but " \
            f"{self.chem_data.number_of_reactions()} reactions are present."

        assert len(conc_arr_before) == self.chem_data.number_of_chemicals(), \
            f"stoichiometry_checker() : the argument `conc_arr_before` must be a Numpy array " \
            f"with as many elements as the number of registered chemicals ({self.chem_data.number_of_chemicals()}); " \
            f"instead, it has {len(conc_arr_before)}"

        assert len(conc_arr_after) == self.chem_data.number_of_chemicals(), \
            f"stoichiometry_checker() : the argument `conc_arr_after` must be a Numpy array " \
            f"with as many elements as the number of registered chemicals ({self.chem_data.number_of_chemicals()}); " \
            f"instead, it has {len(conc_arr_after)}"

        self.chem_data.assert_valid_rxn_index(rxn_index)

        return self._stoichiometry_checker_from_deltas(rxn_index = rxn_index,
                                                       delta_arr = conc_arr_after-conc_arr_before)



    def _stoichiometry_checker_from_deltas(self, rxn_index :int, delta_arr: np.array) -> bool:
        """
        Helper function.

        For the indicated reaction, investigate the change in the concentration of the involved chemicals,
        to ascertain whether the change is consistent with the reaction's stoichiometry.
        See https://life123.science/reactions

        IMPORTANT: this function is currently meant for simulations involving only 1 reaction (TODO: generalize)

        NOTE: the concentration changes in chemicals not involved in the specified reaction are ignored

        :param rxn_index:   The integer index (0-based) to identify the reaction of interest
        :param delta_arr:   Numpy array of numbers, with the concentrations changes of ALL the chemicals (whether involved
                                in the reaction or not), in their index order,
                                as a result of JUST the reaction of interest
        :return:            True if the change in reactant/product concentrations is consistent with the
                                reaction's stoichiometry, or False otherwise
                                Note: if any of the elements of the passed Numpy array is NaN, then True is returned
                                      (because NaN values are indicative of aborted steps; can't invalidate the stoichiometry
                                      check because of that)
        """
        #print("delta_arr: ", delta_arr)

        if np.isnan(delta_arr).any():
            return True         # The presence of a NaN, anywhere in delta_arr, is indicative of an aborted step

        rxn = self.chem_data.get_reaction(rxn_index)
        reactants = self.chem_data.get_reactants(rxn_index)
        products = self.chem_data.get_products(rxn_index)

        # Pick (arbitrarily) the first reactant,
        # to establish a baseline change in concentration relative to its stoichiometric coefficient
        baseline_term = reactants[0]
        baseline_species_name = rxn.extract_species_name(baseline_term)
        baseline_species_index = self.chem_data.get_index(baseline_species_name)
        baseline_stoichiometry = rxn.extract_stoichiometry(baseline_term)
        baseline_ratio =  (delta_arr[baseline_species_index]) / baseline_stoichiometry
        #print("\nbaseline_ratio: ", baseline_ratio)

        for i, term in enumerate(reactants):
            if i != 0:
                species_name = rxn.extract_species_name(term)
                species_index = self.chem_data.get_index(species_name)
                stoichiometry = rxn.extract_stoichiometry(term)
                ratio =  (delta_arr[species_index]) / stoichiometry
                #print(f"ratio for `{self.chem_data.get_name(species)}`: {ratio}")
                if not np.allclose(ratio, baseline_ratio):
                    return False

        for term in products:
            species_name = rxn.extract_species_name(term)
            species_index = self.chem_data.get_index(species_name)
            stoichiometry = rxn.extract_stoichiometry(term)
            ratio =  - (delta_arr[species_index]) / stoichiometry     # The minus in front is b/c we're on the other side of the eqn
            #print(f"ratio for `{self.chem_data.get_name(species)}`: {ratio}")
            if not np.allclose(ratio, baseline_ratio):
                return False

        return True



    def stoichiometry_checker_entire_run(self) -> bool:
        """
        Verify that the stoichiometry is satisfied in all the reaction steps,
        using the diagnostic data from an earlier run

        IMPORTANT: this function is currently meant for simulations involving only 1 reaction (TODO: generalize)

        :return:    True if everything checks out, or False otherwise
        """
        # TODO: maybe make it also accept as inputs a reaction object and a Pandas dataframe
        if self.diagnostic_rxn_data == {}:
            print("WARNING *** In order to run stoichiometry_checker_entire_run(), "
                  "the diagnostics must be turned enabled prior to the simulation run!")
            return False

        number_rxns = self.chem_data.number_of_reactions()
        assert number_rxns == 1, \
            f"stoichiometry_checker_entire_run(): this function is currently designed for just 1 reaction " \
            f"(whereas {number_rxns} are present)"

        for rxn_index in range(number_rxns):
            diagnostic_data = self.get_rxn_data(rxn_index=rxn_index, print_reaction=False)
            for row_index in range(len(diagnostic_data)):
                df_row = diagnostic_data.loc[row_index]     # Row in the Panda's data frame of diagnostic data
                chemical_delta_list = self._delta_names()           # EXAMPLE: ["Delta A", "Delta B", "Delta C"]
                delta = df_row[chemical_delta_list].to_numpy(dtype='float32')      # Extract select columns from the data frame row, and turn into Numpy array
                status = self._stoichiometry_checker_from_deltas(rxn_index=rxn_index, delta_arr=delta)
                if not status:
                    print(f"Stoichiometry NOT satisfied by reaction # {rxn_index}: "
                          f"see row # {row_index} in the diagnostic data for that reaction, from get_diagnostic_rxn_data()")
                    print(df_row )
                    return False

        return True



    def _delta_names(self) -> [str]:
        """
        Return a list of strings, with the names of ALL the registered chemicals,
        in their index order, each prefixed by the string "Delta "
        EXAMPLE: ["Delta A", "Delta B", "Delta X"]

        :return:    A list of strings
        """
        chemical_list = self.chem_data.get_all_labels()      # EXAMPLE: ["A", "B", "X"]
        chemical_delta_list = ["Delta " + name
                               for name in chemical_list]
        return chemical_delta_list



    def _delta_conc_dict(self, delta_conc_arr :np.ndarray) -> dict:
        """
        Convert a Numpy array into a dict, based on all the registered chemicals.
        The keys are the chemical names, prefixed by "Delta "

        :param delta_conc_arr:  A Numpy array of "delta concentrations".  EXAMPLE: array[1.23, 52.2]
        :return:                A dictionary such as {"Delta A": 1.23, "Delta X": 52.2}
        """
        chemical_delta_list = self._delta_names()    # EXAMPLE: ["Delta A", "Delta X"]
        assert len(chemical_delta_list) == len(delta_conc_arr), \
            f"_delta_conc_dict(): mismatch in number of chemicals ({len(chemical_delta_list)} " \
            f"vs. {len(delta_conc_arr)})"

        delta_conc_dict = {}
        for i, delta_name in enumerate(chemical_delta_list):
            delta_conc_dict[delta_name] = delta_conc_arr[i]

        return delta_conc_dict
                            


# General numerical methods

from typing import Union, List, Tuple
import numpy as np
import pandas as pd
import math


class Numerical:
    """
    Assorted, general numerical methods
    """


    @classmethod
    def simple_least_square(cls, X :np.ndarray, Y: np.ndarray) -> (float, float):
        """
        Give two numeric 1-D arrays X and Y, of the same size,
        do a least-square fit: Y = a + b * X , for some numbers a and b

        :param X:   A numpy numeric 1-D array
        :param Y:   A numpy numeric 1-D array, with the same number of elements as X
        :return:    A pair of numbers (a, b) that provide the least-square fit for Y = a + b * X
        """
        assert np.ndim(X) == 1, "simple_least_square(): the argument `X` must be a 1-D array"
        assert np.ndim(Y) == 1, "simple_least_square(): the argument `Y` must be a 1-D array"
        assert len(X) == len(Y), \
            "simple_least_square(): the two arguments must be numeric 1-D arrays with the SAME dimension"
        #TODO: more validations

        M = np.vstack([np.ones(len(X)), X]).T
        # M will be an n x 2 matrix , where n is the number of data points in each of the arguments.
        # The 1st column is all 1's ; the 2nd column contains the values of X
        # EXAMPLE  of M, if X = array([40., 35., 19.]):
        '''
        array([[ 1.        , 40.],
               [ 1.        , 35.],
               [ 1.        , 19.]])
        '''
        a, b = np.linalg.lstsq(M, Y, rcond=None)[0]     # Carry out the least-square fit  as: Y = a + b X
        return a, b


    @classmethod
    def two_vector_least_square(cls, V :np.ndarray, W :np.ndarray, Y :np.ndarray) -> (float, float):
        """
        Give three numeric 1-D arrays V, W and Y, of the same size,
        do a least-square fit: Y = a * V + b * W, for some numbers a and b

        :param V:
        :param W:
        :param Y:
        :return:
        """
        M = np.vstack([V, W]).T
        a, b = np.linalg.lstsq(M, Y, rcond=None)[0]
        return a, b



    @classmethod
    def reach_threshold(cls, df, x :str, y :str, y_threshold) -> Union[float, None]:
        """
        Given a set of 2-D points, whose x- and y- coordinates are stored, respectively, in the
        columns whose names are given by the arguments x and y, locate the x-coordinate at which
        the curve first reaches the given y-value threshold.
        That is, its intersection with the horizontal line y = y_threshold
        Linear interpolation is used.

        The dataframe is *assumed* sorted by the x variable, increasing monotonically; an Exception is
        raised if that's not the case.
        
        Note that if there are multiple intersections, only the leftmost (smallest x) is returned.

        In case no intersection is found, None is returned.

        :param df:          A Pandas dataframe with at least 2 columns
        :param x:           The name of the column with the x-values
        :param y:           The name of the column with the y-values
        :param y_threshold: A number with a y-value being sought in the xy curve
        :return:            The x-value at which the linearly-interpolated curve first reaches
                                the y-value y_threshold, or None if not found
        """

        # Iterate through the DataFrame rows to find the (first) segment where y0 lies between
        for i in range(len(df) - 1):
            # For each row, except the last one

            x1 = df.loc[i, x]
            x2 = df.loc[i + 1, x]
            assert x2 > x1, \
                f"reach_threshold(): the x-values (column `{x}`) are not monotonically-increasing, as expected"

            y1 = df.loc[i, y]                   # y-value at the start of the sliding interval
            y2 = df.loc[i + 1, y]               # y-value at the end of the sliding interval
            if min(y1, y2) <= y_threshold <= max(y1, y2):    # This checks if y0 is between y1 and y2

                if np.allclose(y1, y2):
                    return x1                   # A convention we're following : "when is the target value first reached?"

                # Perform linear interpolation
                x_intersection = x1 + (y_threshold - y1) * (x2 - x1) / (y2 - y1)
                return x_intersection

        return None  # Return None if no intersection is found



    @classmethod
    def curve_intersect(cls, df: pd.DataFrame,
                        x :str, var1 :str, var2 :str, explain=False) -> (float, float):
        """
        Determine the intersection point between 2 Pandas dataframe columns,
        using linear interpolation.
        If more than one is present, only the first (smallest value of x-coord) one will be returned;
        so, the specified time interval should be narrow enough to bracket the intersection of interest

        :param df:      A Pandas dataframe with at least 3 columns
        :param x:       The name of the dataframe column with the independent variable
        :param var1:    The name of the dataframe column with the data from the 1st curve
        :param var2:    The name of the dataframe column with the data from the 2nd curve
        :param explain: [OPTIONAL] If True, print out some details of the computation
        :return:        The pair (x-coordinate of intersection, common value)

        :return:        The pair (x-coordinate of intersection, common value)
        """
        df_size = len(df)   # Number of rows in the dataframe

        # Validation
        assert df_size >= 2, \
            f"curve_intersect(): the Pandas dataframe must have at least 2 rows (it has {df_size}). Not enough data..."

        col_names = list(df.columns)
        assert x in col_names, f"curve_intersect(): the given Pandas dataframe lacks a column named `{x}`"
        assert var1 in col_names, f"curve_intersect(): the given Pandas dataframe lacks a column named `{var1}`"
        assert var2 in col_names, f"curve_intersect(): the given Pandas dataframe lacks a column named `{var2}`"

        # The next 2 integers are the indices of the boundaries of the passed dataframe
        first_index = df.index[0]
        last_index = df.index[-1]
        if explain:
            print(f"first_index: {first_index}, last_index: {last_index}")

        # Examine the y-values of the two curves at the first (zero-th) point
        prev_val1 = df.loc[first_index, var1]
        prev_val2 = df.loc[first_index, var2]

        if np.allclose(prev_val1, prev_val2):     # Trivial intersection immediately found
            if explain:
                print(f"curve_intersect(): Curve intersection found at data row {first_index}, when `{x}` = {df.loc[first_index, x]}")
                if np.allclose(df.loc[1, var1], df.loc[1, var2]):
                    print(f"curve_intersect(): NOTICE - the intersection isn't well-defined because there's extended overlap "
                          f"of the two curves  between `{x}` = {df.loc[first_index, x]} and `{x}` = {df.loc[1, x]}")

            return (df.loc[first_index, x], prev_val1)


        # Make a note of which curve has a bigger value at the initial point
        if prev_val1 > prev_val2:
            initially_larger = 1
        else:
            initially_larger = 2

        row_index_curve_crossover = None

        # Iterate through the DataFrame rows to find the first occurrence of a curve crossover
        # (inversion about which one is larger)
        for i in range(first_index+1, last_index+1):
            # For each row, except the first one

            # y-coords of the two curves at the i-th point
            prev_val1 = df.loc[i, var1]
            prev_val2 = df.loc[i, var2]

            #print(f"val1: {val1} , val2: {val2}")

            if np.allclose(prev_val1, prev_val2):  # Trivial intersection found at the i-th point
                if explain:
                    print(f"curve_intersect(): Curve intersection found at data row {i}, when `{x}` = {df.loc[i, x]}")
                    if (i+1 < df_size) and np.allclose(df.loc[i+1, var1], df.loc[i+1, var2]):
                        print(f"curve_intersect(): NOTICE - the intersection isn't well-defined because there's extended overlap "
                              f"of the two curves between `{x}` = {df.loc[i, x]} and `{x}` = {df.loc[i+1, x]}")

                return (df.loc[i, x], df.loc[i, var1])


            if (prev_val1 > prev_val2) and (initially_larger == 2):
                # Curve 1 is now the larger one
                row_index_curve_crossover = i       # Value inversion found
                break

            if (prev_val2 > prev_val1) and (initially_larger == 1):
                # Curve 2 is now the larger one
                row_index_curve_crossover = i       # Value inversion found
                break
        # END for


        if row_index_curve_crossover is None:      # No crossover found between the two curves
            if explain:
                print("curve_intersect(): No curve crossover detected; i.e., one curve is always above the other one")
            return None         # No intersection located

        if explain:
            print(f"curve_intersect(): Curve crossover detected at data row: {row_index_curve_crossover}")

        before_index = row_index_curve_crossover - 1    # Dataframe row index just before the crossover
        after_index = row_index_curve_crossover         # Dataframe row index just after the crossover


        row_before = df.loc[before_index]
        prev_t = row_before[x]
        prev_val1 = row_before[var1]
        prev_val2 = row_before[var2]

        row_after = df.loc[after_index]
        next_t = row_after[x]
        next_val1 = row_after[var1]
        next_val2 = row_after[var2]

        if np.allclose(prev_t, next_t):     # Curve discontinuity that requires special handling
            print(f"WARNING: curve_intersect(): Discontinuity detected at `{x}` = {prev_t}")
            if np.allclose(prev_val1, next_val1):
                return (prev_t, prev_val1)
            if np.allclose(prev_val2, next_val2):
                return (prev_t, prev_val2)
            return None


        return cls.segment_intersect(  t=(prev_t, next_t),
                                       y1=(prev_val1, next_val1),
                                       y2=(prev_val2, next_val2)
                                    )



    @classmethod
    def segment_intersect(cls, t, y1, y2) -> (float, float):
        """
        Find the intersection of two segments in 2D.
        Their respective endpoints share the same x-coordinates: (t_start, t_end)

        Note: for an alternate method, see line_intersect()

        :param t:   Pair with the joint x-coordinates of the 2 start points,
                        and the joint x-coordinates of the 2 end points
        :param y1:  Pair with the y_coordinates of the endpoints of the 1st segment
        :param y2:  Pair with the y_coordinates of the endpoints of the 2nd segment
                        Note: either segment - but not both - may be horizontal

        :return:    A pair with the (x,y) coord of the intersection;
                        if the segments don't meet their specs (which might result in the lack of an intersection),
                        an Exception is raised
        """
        t_start, t_end = t
        y1_start, y1_end = y1
        y2_start, y2_end = y2

        assert t_end > t_start, \
            f"segment_intersect(): t_end must be strictly > t_start.  Here, t_start = {t_start} , t_end = {t_end}"

        if np.allclose(y1_start, y1_end):       # If line1 is horizontal
            assert y2_start != y2_end, "segment_intersect(): cannot intersect 2 horizontal segments"

        if np.allclose(y2_start, y2_end):       # If line2 is horizontal
            assert y1_start != y1_end, "segment_intersect(): cannot intersect 2 horizontal segments"

        delta_t = t_end - t_start
        delta_r = y1_end - y1_start
        delta_d = y2_end - y2_start

        '''
        Solving for t_intersect the equation:
        (t_intersect - t_start) * (delta_r/delta_t) + y1_start 
            = (t_intersect - t_start) * (delta_d/delta_t) + y2_start
        '''
        factor = (y1_start - y2_start) / (delta_r - delta_d)
        t_intersect = t_start -  factor * delta_t
        y_intersect = y1_start - factor * delta_r

        return t_intersect, y_intersect



    @classmethod
    def line_intersect(cls, p1, p2, q1, q2) -> Union[tuple, None]:
        """
        Returns the coordinates of the intersection
        between the line passing thru the points p1 and p2,
        and the line passing thru the points q1 and q2.

        Based on https://stackoverflow.com/a/42727584/5478830

        Note: for an alternate method, see segment_intersect()

        :param p1:  Pair of (x,y) coord of a point on the 1st line
        :param p2:  Pair of (x,y) coord of another point on the 1st line
        :param q1:  Pair of (x,y) coord of a point on the 2nd line
        :param q2:  Pair of (x,y) coord of another point on the 2nd line

        :return:    A pair with the (x,y) coord of the intersection, if it exists;
                        or None if it doesn't exist
        """
        s = np.vstack([p1, p2, q1, q2])     # s for "stacked"
                                            #   This will be a matrix :
                                            #   [[p1_x, p1_y]
                                            #    [p2_x, p2_y]
                                            #    [q1_x, q1_y]
                                            #    [q2_x, q2_y]]

        vertical_vector = np.ones((4, 1))   # Four 1's stacked up vertically
        mat = np.hstack((s, vertical_vector)) # This is the earlier "s" matrix, with an extra columns of 1's
        l1 = np.cross(mat[0], mat[1])       # This is a vector representation of the 1st line
        l2 = np.cross(mat[2], mat[3])       # This is a vector representation of the 2nd line
        x, y, z = np.cross(l1, l2)          # Data about the point of intersection

        if np.allclose(z, 0.):              # Lines are parallel, or almost so
            return None

        return (x/z, y/z)



    @classmethod
    def deep_flatten(cls, items) -> list:
        """
        Completely flatten any lists/tuples of lists or tuples
        into a single list
        EXAMPLES:
            [[1,2,3], [4,5,6], [7], [8,9]] becomes [1, 2, 3, 4, 5, 6, 7, 8, 9]
            (1,2)                       becomes [1,2]
            [(1, 2), (3,4)]             becomes [1, 2, 3, 4]
            5                           becomes [5]
            "hello"                     becomes ["hello"]

        :param items:   A list or tuple possibly containing lists or tuples
        :return:        A flat list
        """
        if items == [] or items == ():
            return []
        elif (type(items) is not list) and (type(items) is not tuple):
            return [items]
        else:
            return cls.deep_flatten(items[0]) + cls.deep_flatten(items[1:])



    @classmethod
    def compare_results(cls, res1, res2) -> float:
        """
        Use a Euclidean distance metric to compare 2 sets of results - typically from 2 runs
        of different accuracy.
        Each result value may be a number or a list or tuple, possibly containing lists or tuples,
        of numbers.
        However, after the structure is flattened,
        the 2 result data sets must have the same number of points

        :param res1:
        :param res2:
        :return:    The distance between the two sets of value, based on an L2 (Euclidean) metric
        """
        flat1 =  cls.deep_flatten(res1)
        flat2 =  cls.deep_flatten(res2)
        assert len(flat1) == len(flat2), \
            "compare_results(): the 2 sets of data (after flattening) must have the same length"

        return cls.compare_vectors(v1=np.array(flat1), v2=np.array(flat2))



    @classmethod
    def compare_vectors(cls, v1: np.array, v2: np.array, metric=None, trim_edges=0) -> float:
        """
        Return the Euclidean distance between the two given same-sized Numpy arrays
        (TODO: offer option to use alternate metrics)

        :param v1:
        :param v2:
        :param metric:      NOT YET USED: the default L2 norm is always used
        :param trim_edges:  (OPTIONAL) number of elements to ditch at each end,
                                prior to comparing the vectors (default: 0)
        :return:            The distance between the two vectors based on an L2 (Euclidean) metric
        """
        # TODO: give a helpful error message if v1 or v2 aren't 1-d,
        #       or if they have different dimensions
        if trim_edges != 0:
            v1 = v1[trim_edges : -trim_edges]
            v2 = v2[trim_edges : -trim_edges]

        diff = v1 - v2
        return np.linalg.norm(diff, ord=None)



    @classmethod
    def compare_states(cls, state1: np.array, state2: np.array, verbose=False) -> None:
        """
        Print out various assessments of how similar two same-sized Numpy arrays are.
        Typically, those arrays will be system state snapshots - but could be anything.

        :param state1:
        :param state2:
        :param verbose: If True, additional output is shown
        :return:        None
        """
        diff = state1 - state2
        abs_diff = abs(diff)
        print("Max of unsigned absolute differences: ", np.max(abs_diff))
        print("L2 norm of differences (vectors) / Frobenius norm (matrices): ", np.linalg.norm(diff, ord=None))

        rel_diff = diff / state1

        if verbose:
            print("Relative differences: ", rel_diff)

        print("Max of unsigned relative differences: ", np.max(abs(rel_diff)))

        print("Mean of relative differences: ", np.mean(rel_diff))
        print("Median of relative differences: ", np.median(rel_diff))
        print("Standard deviation of relative differences: ", np.std(rel_diff))

        print("np.allclose with lax tolerance?  (rtol=1e-01, atol=1e-01) : ",
              np.allclose(state1 , state2, rtol=1e-01, atol=1e-01))
        print("np.allclose with mid tolerance?  (rtol=1e-02, atol=1e-03) : ",
              np.allclose(state1 , state2, rtol=1e-02, atol=1e-03))
        print("np.allclose with tight tolerance?  (rtol=1e-03, atol=1e-05) : ",
              np.allclose(state1 , state2, rtol=1e-03, atol=1e-05))
        print("np.allclose with extra-tight tolerance?  (rtol=1e-05, atol=1e-08) : ",
              np.allclose(state1 , state2, rtol=1e-05, atol=1e-08))



    @classmethod
    def gradient_order4_1d(cls, arr :Union[np.array, List, Tuple], dx=1.0, dtype='float') -> np.array:
        """
        Compute the gradient, from UNIT-SPACED x-values and y-values in the given 1-dimensional array,
        using the 5-point Central Difference, which produces an accuracy of order 4.

        At the boundary, or close to it, different 5-point stencils are used,
        still resulting in an accuracy of order 4.
        For the coefficients, see: https://web.media.mit.edu/~crtaylor/calculator.html

        Returns the same size as the input array.

        Note:
            * For multi-dimensional cases, use gradient_order4()
            * For a simpler, but less accurate (order 2) approach, use Numpy gradient(),
              which is also usable in case of uneven grids of x values

        :param arr:     One-dimensional Numpy array (or list, or tuple) of numbers,
                            with at least 5 elements
        :param dx:      Delta_x (assumed constant)
        :param dtype:   Data type to use for the elements of the returned array

        :return:        A Numpy array with the same size as the one passed in arr
        """
        arr_size = len(arr)

        assert arr_size >= 5, "gradient_order4_1d(): argument `arr` must have at least 5 elements"

        result = np.zeros(arr_size, dtype=dtype)    # The result is the same size as the input array


        # First, start with the 2 leftmost boundary points

        # For the leftmost boundary point, use the 5-point forward difference
        i = 0
        result[i] = (-25*arr[i] +48*arr[i+1] -36*arr[i+2] +16*arr[i+3] -3*arr[i+4])/12

        # For the 2nd leftmost boundary point, use the skewed 5-point central differences
        i = 1
        result[i] = (-3*arr[i-1] -10*arr[i] +18*arr[i+1] -6*arr[i+2] +arr[i+3])/12


        # Now, process the interior points

        # Coefficients for the 2nd order central finite differences for the first derivative
        C2 = -1/12
        C1 = 2/3

        for i in range(2, arr_size-2):
            result[i] = -C2 * arr[i-2] \
                        -C1 * arr[i-1] \
                        +C1 * arr[i+1] \
                        +C2 * arr[i+2]


        # Finally, conclude with the 2 rightmost boundary points

        # For the 2nd rightmost boundary point, use the skewed 5-point central differences
        i = arr_size-2
        result[i] = (-arr[i-3] +6*arr[i-2] -18*arr[i-1] +10*arr[i] +3*arr[i+1])/12

        # For the rightmost boundary point, use the 5-point backward difference
        i = arr_size-1
        result[i] = (3*arr[i-4] -16*arr[i-3] +36*arr[i-2] -48*arr[i-1] +25*arr[i])/12

        return result / dx



    @classmethod
    def gradient_order4(cls, f: np.array, *varargs) -> Union[np.array, list]:
        """
        Compute the gradient, from the values in the given (possibly multidimensional) array,
        using the 5-point Central Difference, which produces an accuracy of order 4.

        All points need to be equally-spaced along each dimension.

        At the boundary, and at the points immediately adjacent to the boundary,
        the gradient is computed by the very simple 2-point forward or backward differences (accuracy order 1.)

        It returns a list of ndarrays (or a single ndarray if there is only 1 dimension).
        Each list element has the same shape as the original array,
        and corresponds to the derivatives of the passed array with respect to each dimension.

        For 1-dimensional cases, can also use gradient_order4_1d(), which produces accuracy order 4 at ALL points.

        (ADAPTED FROM https://gist.github.com/deeplycloudy/1b9fa46d5290314d9be02a5156b48741 ,
        which is based on 2nd order version from
        http://projects.scipy.org/scipy/numpy/browser/trunk/numpy/lib/function_base.py

        :param f:       An N-dimensional array giving samples of a scalar function
        :param varargs: 0, 1, or N scalars giving the sample distances in each direction

        :return:        A list of N numpy arrays,
                        each of the same shape as f giving the derivative of f with respect to each dimension
        """

        N = len(f.shape)            # Number of dimensions
        n_args = len(varargs)       # Number of arguments passed after the first one (f)
        if n_args == 0:
            dx = [1.0]*N            # List with the element 1.0 repeated N times
        elif n_args == 1:
            dx = [varargs[0]]*N
        elif n_args == N:
            dx = list(varargs)
        else:
            raise SyntaxError(f"gradient_order4(): invalid number of arguments ({n_args})")


        # Use Central Differences on interior, and first differences on endpoints

        out_vals = []       # List of numpy arrays

        # Create slice objects.
        # Initially, they all are "select everything", ie  [:, :, ..., :] across the N dimensions
        # EXAMPLES:
        #           in 1d, the 5 variables below will each be [slice(None, None, None)]
        #           in 2d, each will be [slice(None, None, None), slice(None, None, None)]
        slice0 = [slice(None)]*N
        slice1 = [slice(None)]*N
        slice2 = [slice(None)]*N
        slice3 = [slice(None)]*N
        slice4 = [slice(None)]*N

        out_type = f.dtype.char
        if out_type not in ['f', 'd', 'F', 'D']:   # float or double (real or complex)
            out_type = 'd'      # Use double-precision floats as a default

        for axis in range(N):
            # select out appropriate parts for this dimension
            out = np.zeros(f.shape, dtype=out_type)

            # First, process the Inner Points along this axis (i.e. not the first or last 2)
            slice0[axis] = slice(2, -2)     # Ditch the first 2 and last 2
            slice1[axis] = slice(None, -4)  # Ditch the last 4
            slice2[axis] = slice(1, -3)     # Ditch the 1st and the last 3
            slice3[axis] = slice(3, -1)     # Ditch the first 3 and the last one
            slice4[axis] = slice(4, None)   # Ditch the first 4
            # All the above slice objects will produce a size 4 shorter on the current axis

            # 1D equivalent -- out[2:-2] = (f[:4] - 8*f[1:-3] + 8*f[3:-1] - f[4:])/12.0
            out[slice0] = (f[tuple(slice1)] - 8.0*f[tuple(slice2)] + 8.0*f[tuple(slice3)] - f[tuple(slice4)])/12.0

            # Now process the first 2 points
            slice0[axis] = slice(None, 2)
            slice1[axis] = slice(1, 3)
            slice2[axis] = slice(None, 2)
            # 1D equivalent -- out[0:2] = (f[1:3] - f[0:2])
            out[slice0] = (f[tuple(slice1)] - f[tuple(slice2)])

            # And, finally, the last 3 points
            slice0[axis] = slice(-2, None)
            slice1[axis] = slice(-2, None)
            slice2[axis] = slice(-3, -1)
            # 1D equivalent -- out[-2:] = (f[-2:] - f[-3:-1])
            out[slice0] = (f[tuple(slice1)] - f[tuple(slice2)])


            # Divide by the step size used on this axis
            out_vals.append(out / dx[axis])

            # Reset the slice object in this dimension to ":" (ie, "select everything")
            slice0[axis] = slice(None)
            slice1[axis] = slice(None)
            slice2[axis] = slice(None)
            slice3[axis] = slice(None)
            slice4[axis] = slice(None)
        # END for axis

        if N == 1:
            return out_vals[0]
        else:
            return out_vals



    @classmethod
    def gradient_uneven_grid(cls, x_values :np.array, f :np.array, stencil :int) -> np.ndarray:
        """
        Compute and return the gradient of a 1-D function f(x) at the given grid points,
        which may be uneven.  An arbitrary stencil ("sliding window") size may be used.

        For gradients (first derivatives), the accuracy will be dependent on the size of
        the stencil - though overly large stencils might lead to oscillations.

        Based on B. Fornberg, "Calculation of Weights in Finite Difference Formulas", 1998
        (https://epubs.siam.org/doi/abs/10.1137/S0036144596322507)

        :param x_values:    A Numpy array of x values (independent variable) that
                                may be unevenly-spaced
        :param f:           A Numpy array of the values of the function at the above grid point
        :param stencil:     An integer between 2 and the number of grid points.
        :return:            A Numpy array
        """
        # TODO: allow the gradients to also be computed at x values not on the grid
        assert type(stencil) == int, \
            "gradient_uneven_grid(): value for stencil must be an integer"

        assert stencil >= 2, \
            "gradient_uneven_grid(): value for stencil must be an integer >= 2"

        assert type(x_values) == np.ndarray, \
            f"gradient_uneven_grid(): argument `x_values` must be a Numpy array; instead, it is {type(x_values)}"
        assert type(f) == np.ndarray, \
            f"gradient_uneven_grid(): argument `f` must be a Numpy array; instead, it is {type(f)}"

        assert np.ndim(x_values) == 1, \
            "gradient_uneven_grid(): x_values must be a 1-dimensional Numpy array"
        assert np.ndim(f) == 1, \
            "gradient_uneven_grid(): f must be a 1-dimensional Numpy array"

        size = len(x_values)

        assert size == len(f), \
            "gradient_uneven_grid(): x_values and f must have the same dimension"

        assert stencil <= size, \
            f"gradient_uneven_grid(): value for stencil ({stencil}) cannot exceed the number of grid points ({size})"


        gradient_values = np.zeros(size, dtype=np.float64)

        # Start with the leftmost sliding window, of size stencil, across the array x_values
        # Extract the gradients at point in the left half of that window (rounded up if odd-sized)
        window_x_arr = x_values[ : stencil]   # Array of size equal to stencil value
        assert len(window_x_arr) == stencil
        window_f_arr = f[ : stencil]
        #print("- WINDOW: ", window_x_arr)
        grid_index = 0
        for grid_index in range(math.ceil(stencil / 2)):    # Ceiling of division)
            z = x_values[grid_index]
            #print(f"i : {grid_index} | z : {z}")
            numeric_1st_deriv = cls._compute_derivative(z, window_x_arr, window_f_arr)
            gradient_values[grid_index] = numeric_1st_deriv


        # Advance (by steps of 1) the sliding window across the array x_values, as far as it can go
        for window_start_pos in range(1, size - stencil + 1):
            window_x_arr = x_values[window_start_pos : window_start_pos+stencil]  # Array of size equal to stencil value
            window_f_arr = f[window_start_pos : window_start_pos+stencil]
            assert len(window_x_arr) == stencil
            #print("- WINDOW: ", window_x_arr)
            # Also advance by 1 the grid point at which we compute the gradient
            grid_index += 1
            z = x_values[grid_index]
            #print(f"i : {grid_index} | z : {z}")
            numeric_1st_deriv = cls._compute_derivative(z, window_x_arr, window_f_arr)
            gradient_values[grid_index] = numeric_1st_deriv


        # Process the last remaining grid points, using the last stencil
        # (which is at the end of the x_values array)
        grid_index += 1
        while grid_index < size:    # This loop stops at index size-1
            z = x_values[grid_index]
            #print(f"i : {grid_index} | z : {z}")
            numeric_1st_deriv = cls._compute_derivative(z, window_x_arr, window_f_arr)
            gradient_values[grid_index] = numeric_1st_deriv
            grid_index += 1

        return gradient_values



    @classmethod
    def _compute_derivative(cls, z, window_x_arr, window_f_arr, order=1) -> np.float64:
        """
        Compute and return the specified derivative order (by default the gradient)
        of a 1-D function f(x) at the  point x=z.
        The entire grid of values is used as a stencil.

        :param z:               x value (in grid or not) at which to compute the gradient of f
        :param window_x_arr:    Grid point locations
        :param window_f_arr:    Values of the function at the above grid points
        :param order:           Order of the derivative (by default 1st derivative, i.e. gradient)
        :return:                A floating-point value
        """
        c = cls._finite_diff_weights(z=z, x=window_x_arr, m=order)
        col = c[:, 1]       # Extract 2nd column (k = 1, to be used for 1st derivatives)
        #print("col: ", col)
        #print("window_f_arr: ", window_f_arr)
        numeric_1st_deriv = np.dot(col, window_f_arr)
        return numeric_1st_deriv



    @classmethod
    def _finite_diff_weights(cls, z, x, m) -> np.ndarray:
        """
        Based on B. Fornberg, "Calculation of Weights in Finite Difference Formulas", 1998
        (https://epubs.siam.org/doi/abs/10.1137/S0036144596322507)

        :param z:   x value (in grid or not) where approximations are to be accurate
        :param x:   Grid point locations, x0 thru xn : (n+1) of them
        :param m:   Highest derivative for which weights are sought.
                        EXAMPLE : m=1 to compute up to 1st derivative (gradient)

        :return:    (n+1) x (m+1) Numpy array:
                        c(i, j) stores the weights at grid locations x(i) for derivatives of order j,
                        with 0 <= i <= n , and 0 <= j < m, and n being one less than total number of grid points
        """
        n = len(x) - 1      # One less than total number of grid points

        c = np.zeros((n+1, m+1), dtype=np.float64)

        c1 = 1.0
        c4 = x[0] - z

        c[0, 0] = 1.0

        for i in range(1, n+1):
            mn = min(i, m)
            c2 = 1.0
            c5 = c4
            c4 = x[i] - z

            for j in range(i):
                c3 = x[i] - x[j]
                c2 *= c3

                if j == i - 1:
                    for k in range(mn, 0, -1):
                        c[i, k] = c1 * (k * c[i-1, k-1] - c5 * c[i-1, k]) / c2

                    c[i, 0] = -c1 * c5 * c[i-1, 0] / c2

                for k in range(mn, 0, -1):
                    c[j, k] = (c4 * c[j, k] - k * c[j, k-1]) / c3

                c[j, 0] = c4 * c[j, 0] / c3
            # END for j in range(i)

            c1 = c2

        return c



    @classmethod
    def expand_matrix_boundary(cls, m) -> np.ndarray:
        """
        Add a row at the top and at the bottom, and also add a column to the left and to the right,
        repeating the edge values of the matrix
        EXAMPLE:  [[1, 2],
                   [3, 4]]
                 will turn to
                  [[1, 1, 2, 2],
                   [1, 1, 2, 2],
                   [3, 3, 4, 4],
                   [3, 3, 4, 4]]
                 Note how the original matrix is "embedded" in the center of the larger one

        :param m:   A Numpy matrix
        :return:    A Numpy matrix, with 2 extra rows (at top and bottom) and 2 extra columns (at left and right)
        """
        # Stack up the first row, the matrix, and the last row
        tall = np.concatenate( ([m[0, :]] , m , [m[-1, :]]) )
        #print(tall)

        # Align sideways the 1st column, the matrix and the last column
        wide = np.concatenate( (tall[:,[0]] , tall , tall[:,[-1]]) , axis=1)
        return wide



"""
    life123
    ~~~~~~~

    Dynamical Modeling of Biological Systems

	https://Life123.science

    :copyright: (c) 2022-2024 by Julian West and the Life123 project.
    :license: MIT, see LICENSE for more details.
"""

__version__ = "1.0.0.rc.0"


from life123.bio_sim_1d import BioSim1D
from life123.bio_sim_2d import BioSim2D
from life123.bio_sim_3d import BioSim3D
from life123.chem_data import ChemData
from life123.html_log import HtmlLog
from life123.collections	import (
    CollectionTabular,
    CollectionArray,
    Collection
)
from life123.numerical import Numerical
from life123.reaction import Reaction, ReactionEnz
from life123.thermodynamics import ThermoDynamics
from life123.uniform_compartment import UniformCompartment
from life123.reaction_kinetics import (ReactionKinetics, VariableTimeSteps)

from life123.visualization.graphic_log import GraphicLog
from life123.visualization.plotly_helper import PlotlyHelper
from life123.visualization.py_graph_visual import PyGraphVisual


__all__ = [
    'BioSim1D',
    'BioSim2D',
    'BioSim3D',
    'ChemData',
    'HtmlLog',
    'CollectionTabular',
    'CollectionArray',
    'Collection',
    'Numerical',
    'Reaction',
    'ReactionEnz',
    'ReactionKinetics',
    'ThermoDynamics',
    'VariableTimeSteps',
    'UniformCompartment',
    'GraphicLog',
    'PlotlyHelper',
    'PyGraphVisual'
]



def version():
    return __version__


def check_version(expected :str, enforce=False) -> None:
    """
    Check the passed version number against the actual version number of this library

    :param expected:A string with the expected version number
    :param enforce: If True, a mismatch of versions will result in an Exception
    :return:        None
    """
    if __version__ == expected:
        print("OK")
    else:
        print(f"*** CAUTION: the installed version of the life123 library ({__version__}) "
              f"doesn't match the stated expected version ({expected}).\nIn case of errors, "
              f"change the installed library version, or modify your code to conform to the installed library."
              f"\nChangelog: https://life123.science/history\n\n")

        if enforce:
            raise Exception(f"Using version '{__version__}' instead of the expected version '{expected}'")





from typing import Union, List
from life123.html_log import HtmlLog as log


class GraphicLog:
    """
    To simplify use of HtmlLog and Vue components from within this project
    """

    ###  START OF CONFIGURATION  ###
    #  NOTE: Using local files causes issues in JupyterLab, and it's NOT recommended!
    LOCAL_FILES_ROOT = None     # The project's root folder.  This needs to be set by config().   EXAMPLE: "../../../"
                                # It's the relative path from location of *THE LOG FILE* to Life123's home
    LOCAL_VUE_LIBRARY_FILE = "vue2_lib/vue2.js"                 # Relative to LOCAL_FILES_ROOT
    LOCAL_SVG_HELPER_LIBRARY_FILE = "SVG_helper/svg_helper.js"  # Relative to LOCAL_FILES_ROOT
    LOCAL_VUE_COMPS_DIR = "vue_components/"                     # Relative to LOCAL_FILES_ROOT

    # NOTE: Using remote files is recommended, especially when using JupyterLab
    REMOTE_FILES_ROOT = "https://life123.science/libraries/"
    REMOTE_VUE_LIBRARY_FILE = "https://life123.science/libraries/vue_2.6.12.js"
    REMOTE_SVG_HELPER_LIBRARY_FILE = "https://life123.science/libraries/svg_helper_1.2.js"
    REMOTE_VUE_COMPS_DIR = "https://life123.science/libraries/vue_components/"

    FILES_ROOT = None
    VUE_LIBRARY_FILE = None
    SVG_HELPER_LIBRARY_FILE = None
    VUE_COMPS_DIR = None
    ###  End of configuration  ###


    @classmethod
    def is_initialized(cls) -> bool:
        """
        Return True if this module has been initialized, or False otherwise

        :return:
        """
        if cls.FILES_ROOT is None:    
            return False

        return True



    @classmethod
    def config(cls, filename, components: Union[str, List[str]], mode='overwrite', extra_js=None,
                    local_files=False, home_rel_path=None) -> None:
        """
        Initialize this library
        
        :param filename:        Name, with FULL path, of the desired log file
        :param components:      Either a string or list of strings,
                                with the name(s) of ALL the Vue.js graphic components that will be used in the log file
        :param mode:            A string with the desired logging mode.  Must be one of:
                                    "overwrite" - if the log file exists, first clear it; otherwise, create it
                                    "multiple"  - each run's output should go into a separate file (consecutively numbered)
                                    (note the "append" mode of the class HtmlLog cannot be used)
        :param extra_js:        Optional string with the name of an "extra" JavaScript file to include
                                    ("extra" refers to the fact that several JS files are already automatically included)
                                
        :param local_files:     If True, local files will be used for the Vue components (see arg "home_rel_path")
        :param home_rel_path:   Relative path from location of *THE LOG FILE* to the Life123's project home
                                    NEW: ONLY APPLICABLE IF local_files=True ; otherwise, it's simply ignored
                                    EXAMPLE: "../../.."

        :return:                None
        """
        if local_files:
            assert home_rel_path, \
                "config(): if the local_files=True flag is set, then an argument must be passed for `home_rel_path`"

            cls.FILES_ROOT = home_rel_path
            cls.VUE_LIBRARY_FILE = home_rel_path + "/" + cls.LOCAL_VUE_LIBRARY_FILE
            cls.SVG_HELPER_LIBRARY_FILE = home_rel_path + "/" + cls.LOCAL_SVG_HELPER_LIBRARY_FILE
            cls.VUE_COMPS_DIR = home_rel_path + "/" + cls.LOCAL_VUE_COMPS_DIR
        else:
            assert not home_rel_path, \
                "config(): if the local_files=Flase flag is set, then the `home_rel_path` must be None"

            cls.FILES_ROOT = cls.REMOTE_FILES_ROOT
            cls.VUE_LIBRARY_FILE = cls.REMOTE_VUE_LIBRARY_FILE
            cls.SVG_HELPER_LIBRARY_FILE = cls.REMOTE_SVG_HELPER_LIBRARY_FILE
            cls.VUE_COMPS_DIR = cls.REMOTE_VUE_COMPS_DIR


        if type(components) == str:
            css_files = f"{cls.VUE_COMPS_DIR}{components}.css"
        elif type(components) == list:
            css_files = [f"{cls.VUE_COMPS_DIR}{comp}.css"
                                        for comp in components]
        else:
            raise Exception("GraphicLog.config(): argument `components` must be either a string or a list of strings")


        assert mode == "overwrite" or mode == "append", \
                "GraphicLog.config(): argument `mode` must be either 'overwrite' or 'append'"

        # Assemble a list of all the needed JavaScript files
        js = cls.SVG_HELPER_LIBRARY_FILE
        if extra_js:
            js = [js, extra_js]

        # Note: paths are from the location of *THE LOG FILE*
        log.config(filename=filename, mode=mode,
                   use_D3=True,
                   Vue_lib = cls.VUE_LIBRARY_FILE,
                   js  = js,
                   css = css_files)



    @classmethod
    def export_plot(cls, graph_data: dict, graphic_component: str, print_notification=True, unpack=False) -> None:
        """
        Send to the log file the data to create a Vue-based plot

        :param graph_data:          A python dictionary of data to pass to the Vue component
        :param graphic_component:   A string with the name of the existing Vue.js component to use.
                                        EXAMPLE: "vue_curves_4" (assuming that a js file with such a Vue component exists)
        :param print_notification:  If True, something is printed to inform of what's happening with the log file
        :param unpack:              Use True for Vue components that require their data unpacked into individual arguments;
                                        False for that accept a single data argument, named "graph_data"
        :return:                    None
        """
        assert cls.FILES_ROOT is not None, \
            "export_plot(): must first initialize library with call to GraphicLog.config()"

        if unpack:
            log.export_plot_Vue_unpack_args(data=graph_data,
                                            component_name = graphic_component,
                                            component_file = f"{cls.VUE_COMPS_DIR}{graphic_component}.js")
        else:
            log.export_plot_Vue(graph_data=graph_data,
                                component_name = graphic_component,
                                component_file = f"{cls.VUE_COMPS_DIR}{graphic_component}.js")

        if print_notification:
            print(f"[GRAPHIC ELEMENT SENT TO LOG FILE `{log.log_fullname}`]")

